#!/usr/bin/env ruby
#
#
#

require 'pp'
require 'erb'

unless defined?($srcdir)
  $srcdir = File.dirname($0) + '/..'
end

class InsnsDef
  VARS = {}
  
  def initialize file= $srcdir+'/insns.def', optopfile = $srcdir+'/opt_operand.def'
    @insn  = nil
    @insns = []
    @insn_opes = Hash.new{ [].freeze } # insn => [[type, ope1], ...]
    @insn_pops = Hash.new{ [].freeze } # insn => [[type, pop1], ...]
    @insn_rets = Hash.new{ [].freeze } # insn => [[type, ret1], ...]
    @insn_body = Hash.new{|h, k| h[k] = ''} # insn => '...'
    @insn_comm = Hash.new              # insn => /* ... */
    @insn_type = Hash.new{ nil }       #
    @insns_sc  = Hash.new{ nil }

    # for optimize
    @insn_defopes = Hash.new{ [].freeze }
    @insn_nextsc  = Hash.new{ [].freeze }
    @insn_push    = Hash.new{ [[].freeze, [].freeze].freeze }
    # insn => [[before regs], [after regs]]

    load_insns_def open(file)
    load_opt_operand_def open(optopfile)  
    make_stackcaching_insns if $opts['OPT_STACK_CACHING']
  end

  attr_reader :insns, :insn_opes, :insn_pops, :insn_rets, :insn_comm
  attr_reader :insn_body
  
  SPECIAL_INSN_FOR_SC_AFTER = {
    /\Asend/      => [:a],
    /\Aend/       => [:a],
    /\Ayield/     => [:a],
    /\Aclassdef/  => [:a],
    /\Amoduledef/ => [:a],
  }
  FROM_SC = [[], [:a], [:b], [:a, :b], [:b, :a]]
  
  def make_stackcaching_insns
    pops = rets = nil
    
    @insns.dup.each{|insn|
      opops = @insn_pops[insn]
      orets = @insn_rets[insn]
      oopes = @insn_opes[insn]
      ocomm = @insn_comm[insn]

      after = nil
      SPECIAL_INSN_FOR_SC_AFTER.any?{|k, v|
        if k =~ insn
          after = v
          break
        end
      }
      insns = []
      FROM_SC.each{|from|
        name, pops, rets, pushs1, pushs2, nextsc =
          *calc_stack(insn, from, after, opops, orets)
        
        if true
          make_new_insn(name, oopes, pops, rets, ocomm)
          @insn_push[name]    = [pushs1, pushs2]
        else
          # experiment
          make_new_insn(name, oopes, opops, orets, ocomm)
          # @insn_push[name]    = # [pushs1, pushs2]
        end

        @insn_body[name]    = @insn_body[insn]
        @insn_nextsc[name]  = nextsc
        @insn_defopes[name] = @insn_defopes[insn]
        @insn_type[name]    = :sc
        insns << name
      }
      @insns_sc[insn] = insns
    }
    # exit! 1
  end

  def complement_name st
    "#{st[0] ? st[0] : 'x'}#{st[1] ? st[1] : 'x'}"
  end

  def add_stack_value st
    len = st.length
    if len == 0
      st[0] = :a
      [nil, :a]
    elsif len == 1
      if st[0] == :a
        st[1] = :b
      else
        st[1] = :a
      end
      [nil, st[1]]
    else
      st[0], st[1] = st[1], st[0]
      [st[1], st[1]]
    end
  end

  def calc_stack insn, ofrom, oafter, opops, orets
    from = ofrom.dup; pops = opops.dup; rets = orets.dup
    pushs_before = []
    pushs= []
    
    pops.each_with_index{|e, i|
      if e[0] == '...'
        pushs_before = from
        from = []
      end
      r = from.pop
      break unless r
      pops[i] = pops[i].dup << r
    }

    if oafter
      from = oafter
      from.each_with_index{|r, i|
        rets[i] = rets[i].dup << r if rets[i]
      }
    else
      rets = rets.reverse
      rets.each_with_index{|e, i|
        break if e[0] == '...'
        pushed, r = add_stack_value from
        rets[i] = rets[i].dup << r
        pushs << pushed if pushed
      }
    end
    
    if false # insn == 'swap'
      p ofrom
      p pops
      p rets
      p pushs_before
      p pushs
    end
    
    ret = ["#{insn}_SC_#{complement_name(ofrom)}_#{complement_name(from)}",
            pops, rets, pushs_before, pushs, from]
    #p ret
    ret
  end
  
  # str -> [[type, var], ...]
  def parse_vars line
    raise unless /\((.*)\)/ =~ line
    vars = $1.split(',')
    vars.map!{|v|
      if /\s*(\S+)\s+(\S+)\s*/ =~ v
        type = $1
        var  = $2
      elsif /\s*\.\.\.\s*/ =~ v
        type = var  = '...'
      else
        raise
      end
      [type, var]
    }
    vars
  end

  def parse_comment comm
    c = 'others'
    j = ''
    e = ''
    comm.each{|line|
      case line
      when /@c (.+)/
        c = $1
      when /@e (.+)/
        e = $1
      when /@e\s*$/
        e = ''
      when /@j (.+)$/
        j = $1
      when /@j\s*$/
        j = ''
      end
    }
    { :c => c,
      :e => e,
      :j => j,
    }
  end

  def make_header_prepare_stack insn
    ret = ''
    push_ba = @insn_push[insn]
    raise "unsupport" if push_ba[0].size > 0 && push_ba[1].size > 0
    
    push_ba.each{|pushs|
      pushs.each{|r|
        ret += "PUSH(SCREG(#{r}));\n"
      }
    }
    ret
  end
  
  def make_header_operands insn
    vars = @insn_opes[insn]
    n = 0
    ops = []
    
    vars.each_with_index{|(type, var), i|
      if type == '...'
        break
      end

      ops << "  #{type} #{var} = (#{type})GET_OPERAND(#{i+1});\n"
      n   += 1
    }
    @opn = n
    
    # reverse or not?
    # ops.join
    ops.reverse.join
  end
  
  def make_header_default_operands insn
    ret = ''
    vars = @insn_defopes[insn]
    vars.each{|e|
      next if e[1] == '*'
      ret += "#define #{e[0][1]} #{e[1]}\n"
      # "  const #{e[0][0]} #{e[0][1]} = #{e[1]};\n"
    }
    ret
  end

  def make_footer_default_operands insn
    ret = ''
    vars = @insn_defopes[insn]
    vars.each{|e|
      next if e[1] == '*'
      ret += "#undef #{e[0][1]}\n"
      # "  const #{e[0][0]} #{e[0][1]} = #{e[1]};\n"
    }
    ret
  end
  
  def make_header_stack_pops insn
    n = 0
    pops = []
    
    vars = @insn_pops[insn]
    vars.each_with_index{|(type, var, r), i|
      if type == '...'
        break
      end
      if r
        pops << "  #{type} #{var} = SCREG(#{r});\n"
      else
        pops << "  #{type} #{var} = TOPN(#{n});\n"
        n += 1
      end
    }
    @popn = n

    # reverse or not?
    pops.reverse.join
  end

  def make_header_stack_val insn
    ret = ''
    vars = @insn_opes[insn] + @insn_pops[insn] + @insn_defopes[insn].map{|e| e[0]}
    @insn_rets[insn].each{|var|
      if vars.all?{|e| e[1] != var[1]} && var[1] != '...'
        ret += "  #{var[0]} #{var[1]};\n"
      end
    }
    ret
  end

  def make_footer_stack_val insn
    ret = ''
    @insn_rets[insn].reverse_each{|v|
      if v[1] == '...'
        break
      end
      if v[2]
        ret += "  SCREG(#{v[2]}) = #{v[1]};\n"
      else
        ret += "  PUSH(#{v[1]});\n"
      end
    }
    ret
  end
  
  def make_header insn
    ret  = "\nINSN_ENTRY(#{insn}){\n"
    ret += "  /* prepare stack status */\n"
    ret += make_header_prepare_stack insn
    ret += "{\n"
    ret += "  /* declare stack push val */\n"
    ret += make_header_stack_val  insn
    ret += "  /* declare and initialize default opes */\n"
    ret += make_header_default_operands insn
    ret += "  /* declare and get from iseq */\n"
    ret += make_header_operands   insn
    ret += "  /* declare and pop from stack */\n"
    ret += make_header_stack_pops insn
    
    ret += "  /* for debug */\n"
    ret += "  DEBUG_ENTER_INSN(\"#{insn}\");\n"
    ret += "  USAGE_ANALYSIS_INSN(BIN(#{insn}));\n"
    
    ret += "  /* management */\n"
    ret += "  ADD_PC(1+#{@opn});\n"
    ret += "  POPN(#{@popn});\n" if @popn > 0
    ret += "  #define CURRENT_INSN_#{insn} 1\n"
    ret += "  #define INSN_IS_SC()     #{@insns_sc[insn] ? 0 : 1}\n"
    ret += "  #define INSN_LABEL(lab)  LABEL_#{insn}_##lab\n"
    ret += "  #define LABEL_IS_SC(lab) LABEL_##lab##_###{@insns_sc[insn] ? 't' : 'f'}\n"
    @insn_opes[insn].each_with_index{|op, i|
      ret += "  USAGE_ANALYSIS_OPERAND(BIN(#{insn}), #{i}, #{op[1]});\n"
    }
    ret += "{\n"
    
  end

  def make_footer insn
    ret  = "  /* push stack val */\n"
    ret += make_footer_stack_val insn
    # debug info

    # epilogue
    ret += make_footer_default_operands insn
    ret += "#undef CURRENT_INSN_#{insn}\n"
    ret += "#undef INSN_IS_SC\n"
    ret += "#undef INSN_LABEL\n"
    ret += "#undef LABEL_IS_SC\n"
    ret += "  END_INSN();\n}}\n"
    ret += "}\n"
  end

  def make_body insn
    @insn_body[insn]
  end
  
  def make_insn insn
    ret  = make_header insn
    ret += make_body   insn
    ret += make_footer insn
  end

  def make_new_insn insn, opes, pops, rets, comm
    @insns << insn
    @insn_opes[insn] = opes
    @insn_pops[insn] = pops
    @insn_rets[insn] = rets
    @insn_comm[insn] = comm
  end
  
  def load_insns_def f
    skip_comment_pattern = Regexp.compile(Regexp.escape('/** ##skip'))
    comment = ''
    insn    = nil
    insn_in = false
    
    while line = f.gets
      line.chomp!
      case line
        
      # skip comment
      when skip_comment_pattern
        while line = f.gets.chomp
          if /\s+\*\/$/ =~ line
            break
          end
        end
        
      # collect instruction comment
      when /^\/\*\*$/
        while line = f.gets
          if /\s+\*\/\s*$/ =~ line
            break
          else
            comment << line
          end
        end
        
      # start instruction body
      when /^DEFINE_INSN$/
        insn = f.gets.chomp
        opes = parse_vars(f.gets.chomp)
        pops = parse_vars(f.gets.chomp).reverse
        rets = parse_vars(f.gets.chomp).reverse
        
        @insn = insn
        
        make_new_insn insn, opes, pops, rets, parse_comment(comment)
        comment = ''
        insn_in = true
        
        raise unless /^\{$/ =~ f.gets.chomp

      when /^\}/
        insn_in = false

      # 
      else
        if insn_in
          @insn_body[@insn] << line + "\n"
        end
      end
    end

    if false
      insns = @insns.dup
      10.times{|i|
        insns.each{|insn|
          comm[:c] = 'optimize'
          make_new_insn "#{insn}_dmy_#{i}", # [], [], [], []
          @insn_opes[insn], @insn_pops[insn], @insn_rets[insn], @insn_comm[insn]
          @insn_body["#{insn}_#{i}"] = "/* nop */;\n"
        }
      }
    end
    
  end

  def label_escape label
    label.gsub(/\(/, '_OP_').
          gsub(/\)/, '_CP_').
          gsub(/\*/, '_WC_')
  end
  
  def add_opt_operand insn, opts
    opes = @insn_opes[insn]
    raise "operand size mismatcvh for #{insn} (#{opts.size})" if opes.size != opts.size
    ninsn = insn + '_opopt_' + opts.map{|e| label_escape(e)}.join('_')
    nopes = []
    defv  = []
    
    opts.each_with_index{|e, i|
      if e == '*'
        nopes << opes[i]
      end
      defv  << [opes[i], e]
    }

    comm = @insn_comm[insn]
    comm[:c] = 'optimize'
    
    make_new_insn(ninsn, nopes, @insn_pops[insn], @insn_rets[insn], comm)
    @insn_defopes[ninsn] = defv
    @insn_body[ninsn]    = @insn_body[insn]
  end
  
  def load_opt_operand_def f
    f.each{|line|
      line = line.gsub(/\#.*/, '').strip
      next  if line.length == 0
      break if /__END__/ =~ line
      /(\S+)\s+(.+)/ =~ line
      insn = $1
      opts = $2
      add_opt_operand insn, opts.split(/,/).map{|e| e.strip}
    }
  end
  
  # vm.inc
  def vm_inc
    vars = ''
    vm_body = ''

    @insns.each{|insn|
    #@insns.sort.each{|insn|
      vm_body << "\n"
      vm_body << make_insn(insn)
    }
    
    ERB.new(File.read($srcdir + '/tmpl/vm.inc.tmpl')).result(binding)
  end

  # vmtc.inc
  def vmtc_inc
    insns_table = ''
    @insns.each{|insn|
      insns_table << "  &&LABEL(#{insn}),\n"
    }
    
    ERB.new(File.read($srcdir + '/tmpl/vmtc.inc.tmpl')).result(binding)
  end
  
  def op2typesig op
    case op
    when /^long/
      "TS_LONG"
    when /^ulong/
      "TS_ULONG"
    when /^VALUE/
      "TS_VALUE"
    when /^ID/
      "TS_ID"
    when /GENTRY/
      "TS_GENTRY"
    when /^IC/
      "TS_IC"
    when /^\.\.\./
      "TS_VARIABLE"
    else
      raise
    end
  end

  # insns_info.inc
  def insns_info_inc
    # insn_names
    insn_names = ''
    @insns.each{|insn|
      insn_names << "  \"#{insn}\",\n"
    }

    # operands info
    operands_info = ''
    operands_num_info = ''
    @insns.each{|insn|
      opes = @insn_opes[insn]
      operands_info << '  '
      if opes.size == 0
        operands_info << '"",' + "\n"
      else
        opes.each{|type, var|
          operands_info << (op2typesig(type) + ' ')
        }
        operands_info << ",\n"
      end

      num = opes.size + 1
      operands_num_info << "  #{num},\n"
    }

    # stack num
    stack_num_info = ''
    @insns.each{|insn|
      num = @insn_rets[insn].size
      stack_num_info << "  #{num},\n"
    }
    
    ERB.new(File.read($srcdir + '/tmpl/insns_info.inc.tmpl')).result(binding)
  end

  # insns.inc
  def insns_inc
    insns = ''
    i=0
    @insns.each{|insn|
      insns << "  %-30s = %d,\n" % ["BIN(#{insn})", i]
      i+=1
    }
    ERB.new(File.read($srcdir + '/tmpl/insns.inc.tmpl')).result(binding)
  end

  # minsns.inc
  def minsns_inc
    defs = ''
    i=0
    @insns.each{|insn|
      defs << "  rb_define_const(mYarvInsns, %-30s, INT2FIX(%d));\n" % ["\"I#{insn}\"", i]
      i+=1
    }
    ERB.new(File.read($srcdir + '/tmpl/minsns.inc.tmpl')).result(binding)
  end

  def val_as_type op
    type = op[0][0]
    val  = op[1]

    case type
    when /^long/
      "INT2FIX(#{val})"
    when /^ulong/
      "INT2FIX(#{val})"
    when /^VALUE/
      val
    when /^ID/
      "INT2FIX(#{val})"
    when /GENTRY/
      raise
    when /^\.\.\./
      raise
    else
      raise
    end
    
  end
  
  # optinsn.inc
  def optinsn_inc
    rule = ''
    insns = {}
    @insns.each{|insn|
      next if @insn_defopes[insn].size == 0
      next if @insn_type[insn] == :sc
      
      originsn = /(.+?)_/.match(insn)[1]
      insns[originsn] ? insns[originsn] << insn : insns[originsn] = [insn]
    }
    insns.each{|originsn, optinsns|
      rule += "case BIN(#{originsn}):\n"
      optinsns.sort_by{|opti|
        # p @insn_defopes[opti].find_all{|e|p e; e[1] == '*'}.size
        @insn_defopes[opti].find_all{|e| e[1] == '*'}.size
      }.each{|opti|
        rule += "  if(\n"
        i = 0
        rule += '      ' + @insn_defopes[opti].map{|opinfo|
          i += 1
          next if opinfo[1] == '*'
          ["GET_OPERAND(#{i-1}) == #{val_as_type(opinfo)}\n"]
        }.compact.join('  &&  ')
        rule += "  ){\n"
          idx = 0
          n = 0
          @insn_defopes[opti].each{|opinfo|
            if opinfo[1] == '*'
              idx += 1
            else
              rule += "  rb_ary_delete_at(insnobj->operands, #{idx});\n"
            end
            n += 1
          }
          rule += "  self = new_insn(BIN(#{opti}), insnobj->line_no, insnobj->operands);\n"
        rule += "  break;}\n"
      }
      rule += "break;\n";
    }
    ERB.new(File.read($srcdir + '/tmpl/optinsn.inc.tmpl')).result(binding)
  end

  
  def opt_sc_inc
    sc_insn_info = []
    @insns.each{|insn|
      if insns = @insns_sc[insn]
        insns = ['SC_ERROR'] + insns.map{|e| "    BIN(#{e})"}
      else
        insns = Array.new(6){'SC_ERROR'}
      end
      sc_insn_info << "  {\n#{insns.join(",\n")}}"
    }
    sc_insn_info = sc_insn_info.join(",\n")
    
    sc_insn_next = @insns.map{|insn|
      "  SCS_#{complement_name(@insn_nextsc[insn]).upcase}"
    }.join(",\n")
    ERB.new(File.read($srcdir + '/tmpl/opt_sc.inc.tmpl')).result(binding)
  end
    
  def desc lang
    d = ''
    i = 0
    cat = nil
    @insns.each{|insn|
      seq    = @insn_opes[insn].map{|t,v| v}.join(' ')
      before = @insn_pops[insn].reverse.map{|t,v| v}.join(' ')
      after  = @insn_rets[insn].reverse.map{|t,v| v}.join(' ')

      if cat != @insn_comm[insn][:c]
        cat = @insn_comm[insn][:c]
        d << "** #{cat}\n\n"
      end
      
      d << "*** #{insn}\n"
      d << "\n"
      d << @insn_comm[insn][lang] + "\n\n"
      d << ":instruction sequence: 0x%02x #{seq}\n" % i
      d << ":stack: #{before} => #{after}\n\n"
      i+=1
    }
    d
  end
  
  def desc_ja
    d = desc :j
    ERB.new(File.read($srcdir + '/tmpl/yarvarch.ja')).result(binding)
  end

  def desc_en
    d = desc :e
    ERB.new(File.read($srcdir + '/tmpl/yarvarch.en')).result(binding)
  end

  Files = { # codes
    'vm.inc'         => :vm_inc,
    'vmtc.inc'       => :vmtc_inc,
    'insns.inc'      => :insns_inc,
    'insns_info.inc' => :insns_info_inc,
    'minsns.inc'     => :minsns_inc,
    'optinsn.inc'    => :optinsn_inc,
    'opt_sc.inc'     => :opt_sc_inc,
  }
  
  def self.make_sources args = []
    insns = InsnsDef.new
    args = Files.keys if args.empty?
    
    args.each{|fn|
      s = Files[fn]
      open(fn, 'w'){|f|
        f.puts(insns.__send__(s))
      }
    }
  end
end

files = []
$opts = {}

ARGV.each{|e|
  if /\A\-D(\w+)/ =~ e
    $opts[$1] = true
  elsif /\-/ =~ e
    # ignore
  else
    files << e
  end
}

if $0 == __FILE__
  InsnsDef.make_sources files
end

